import torch
from torch.nn import Module
from torch import nn
import math


class SinusoidalPositionEmbedding(Module):
    """定义Sin-Cos位置Embedding
    """

    def __init__(
            self, output_dim, merge_mode='add', custom_position_ids=False):
        super(SinusoidalPositionEmbedding, self).__init__()
        self.output_dim = output_dim
        self.merge_mode = merge_mode
        self.custom_position_ids = custom_position_ids

    def forward(self, inputs):
        input_shape = inputs.shape
        batch_size, seq_len = input_shape[0], input_shape[1]
        position_ids = torch.arange(seq_len).type(torch.float)[None]
        indices = torch.arange(self.output_dim // 2).type(torch.float)
        indices = torch.pow(10000.0, -2 * indices / self.output_dim)
        embeddings = torch.einsum('bn,d->bnd', position_ids, indices)
        embeddings = torch.stack([torch.sin(embeddings), torch.cos(embeddings)], dim=-1)
        embeddings = torch.reshape(embeddings, (-1, seq_len, self.output_dim))
        if self.merge_mode == 'add':
            return inputs + embeddings.to(inputs.device)
        elif self.merge_mode == 'mul':
            return inputs * (embeddings + 1.0).to(inputs.device)
        elif self.merge_mode == 'zero':
            return embeddings.to(inputs.device)


def relative_position_encoding(depth, max_length=512, max_relative_position=127):
    vocab_size = max_relative_position * 2 + 1
    range_vec = torch.arange(max_length)
    range_mat = range_vec.repeat(max_length).view(max_length, max_length)
    distance_mat = range_mat - torch.t(range_mat)
    distance_mat_clipped = torch.clamp(distance_mat, -max_relative_position, max_relative_position)
    final_mat = distance_mat_clipped + max_relative_position

    embeddings_table = torch.zeros(vocab_size, depth)
    position = torch.arange(0, vocab_size, dtype=torch.float).unsqueeze(1)
    div_term = torch.exp(torch.arange(0, depth, 2).float() * (-math.log(10000.0) / depth))
    embeddings_table[:, 0::2] = torch.sin(position * div_term)
    embeddings_table[:, 1::2] = torch.cos(position * div_term)
    embeddings_table = embeddings_table.unsqueeze(0).transpose(0, 1).squeeze(1)

    flat_relative_positions_matrix = final_mat.view(-1)
    one_hot_relative_positions_matrix = torch.nn.functional.one_hot(flat_relative_positions_matrix,
                                                                    num_classes=vocab_size).float()
    positions_encoding = torch.matmul(one_hot_relative_positions_matrix, embeddings_table)
    my_shape = list(final_mat.size())
    my_shape.append(depth)
    positions_encoding = positions_encoding.view(my_shape)
    return positions_encoding


def sequence_masking(x, mask, value='-inf', axis=None):
    if mask is None:
        return x
    else:
        if value == '-inf':
            value = -1e12
        elif value == 'inf':
            value = 1e12
        assert axis > 0, 'axis must be greater than 0'
        for _ in range(axis - 1):
            mask = torch.unsqueeze(mask, 1)
        for _ in range(x.ndim - mask.ndim):
            mask = torch.unsqueeze(mask, mask.ndim)
        return x * mask + value * (1 - mask)


def add_mask_tril(logits, mask):
    if mask.dtype != logits.dtype:
        mask = mask.type(logits.dtype)
    logits = sequence_masking(logits, mask, '-inf', logits.ndim - 2)
    logits = sequence_masking(logits, mask, '-inf', logits.ndim - 1)
    # 排除下三角
    mask = torch.tril(torch.ones_like(logits), diagonal=-1)
    logits = logits - mask * 1e12
    return logits


class GlobalPointer(Module):
    """全局指针模块
    将序列的每个(start, end)作为整体来进行判断
    """
    # last_hidden_state (torch.FloatTensor of shape (batch_size, sequence_length, hidden_size))
    #(c_size, head_size, embedding_size) 9 64 1024
    def __init__(self, heads, head_size, hidden_size, RoPE=True):
        super(GlobalPointer, self).__init__()
        self.heads = heads
        self.head_size = head_size
        self.RoPE = RoPE
        self.dense = nn.Linear(hidden_size, self.head_size * self.heads * 2)

    def forward(self, inputs, mask=None):
        # last_hidden_state:(batch_size, seq_len, hidden_size)

        # outputs:(batch_size, seq_len, ent_type_size*inner_dim*2)      ent_type=heads  inner_dim = head_size
        inputs = self.dense(inputs)
        inputs = torch.split(inputs, self.head_size * 2, dim=-1)

        # 按照-1这个维度去分，每块包含x个小块
        # outputs:(batch_size, seq_len, ent_type_size, inner_dim*2)
        inputs = torch.stack(inputs, dim=-2)

        # 沿着一个新维度对输入张量序列进行连接。 序列中所有的张量都应该为相同形状
        # qw,kw:(batch_size, seq_len, ent_type_size, inner_dim)
        qw, kw = inputs[..., :self.head_size], inputs[..., self.head_size:]
        # 分出qw和kw
        # RoPE编码
        if self.RoPE:
            # pos_emb:(batch_size, seq_len, inner_dim)
            pos = SinusoidalPositionEmbedding(self.head_size, 'zero')(inputs)
            # cos_pos,sin_pos: (batch_size, seq_len, 1, inner_dim)
            cos_pos = pos[..., None, 1::2].repeat(1, 1, 1, 2)
            sin_pos = pos[..., None, ::2].repeat(1, 1, 1, 2)
            qw2 = torch.stack([-qw[..., 1::2], qw[..., ::2]], 4)
            qw2 = torch.reshape(qw2, qw.shape)
            qw = qw * cos_pos + qw2 * sin_pos
            kw2 = torch.stack([-kw[..., 1::2], kw[..., ::2]], 4)
            kw2 = torch.reshape(kw2, kw.shape)
            kw = kw * cos_pos + kw2 * sin_pos
        # 计算内积
        # logits:(batch_size, ent_type_size, seq_len, seq_len)
        logits = torch.einsum('bmhd , bnhd -> bhmn', qw, kw)
        # 排除padding 排除下三角
        logits = add_mask_tril(logits, mask)
        # scale返回
        return logits / self.head_size ** 0.5



import os
import numpy as np
import json
import torch
from torch import nn
from torch.utils.data import Dataset, DataLoader
from transformers import BertModel, BertTokenizerFast
from .model import GlobalPointer
from .utils import rematch

__all__=['load_NNER']


class _CustomDataset4test(Dataset):
    def __init__(self, data, tokenizer, maxlen, c_size):
        self.data = data
        self.tokenizer = tokenizer
        self.maxlen = maxlen
        self.c_size = c_size

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        d = self.data[idx]
        label = torch.zeros((self.c_size, self.maxlen, self.maxlen))
        enc_context = self.tokenizer(d[0], return_offsets_mapping=True, max_length=self.maxlen, truncation=True,
                                padding='max_length', return_tensors='pt')
                                
        enc_context = {key: enc_context[key][0] for key in enc_context.keys() if enc_context[key].shape[0] == 1}

        return enc_context

class _GlobalPointerNet(nn.Module):
    def __init__(self, model_name, c_size=9, head_size=64, embedding_size=1024):
        super(_GlobalPointerNet, self).__init__()
        self.head = GlobalPointer(c_size, head_size, embedding_size)
        self.bert = BertModel.from_pretrained(model_name)

    def forward(self, input_ids, attention_mask, token_type_ids):
        x1 = self.bert(input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
        x2 = x1.last_hidden_state
        #last_hidden_state (torch.FloatTensor of shape (batch_size, sequence_length, hidden_size))
        # – Sequence of hidden-states at the output of the last layer of the model.

        logits = self.head(x2, mask=attention_mask)
        return logits

class _NamedEntityRecognizer(object):
    """
    Named Entity Recognizer
    """
    def __init__(self, maxlen, c_size, id2c, c2c, device, tokenizer, model_base):
        self.maxlen = maxlen
        self.c_size = c_size
        self.id2c = id2c
        self.c2c = c2c
        self.device = device
        self.tokenizer = tokenizer
        self.model_base = model_base

    def recognize(self, text, threshold=0):
        """
        Args:
            text (str): input sentence
            threshold (int, optional): threshold score to filter out recognized entity with low confidence score
                Default: 0
        Returns:
            entities (list): list of dicts with recognized entity.
                The keys are: start_idx, end_idx, type, Chinese_type, entity
        """
        self.model_base.eval()
        text = text[0:self.maxlen]
 
        tokens = self.tokenizer.tokenize(text)
        mapping = rematch(self.tokenizer, text, tokens)
        text = [[str(text)]]
        data = _CustomDataset4test(text,self.tokenizer,self.maxlen,self.c_size)
        data = DataLoader(data)
        input_ids = []
        attention_mask = []
        token_type_ids = []
        
        for data_1 in data:
            input_ids = data_1['input_ids'].to(self.device)
            attention_mask = data_1['attention_mask'].to(self.device)
            token_type_ids = data_1['token_type_ids'].to(self.device)


        with torch.no_grad():
            scores = self.model_base(input_ids, attention_mask, token_type_ids)

        entities = []
        lenss = len(mapping)

        
        score = scores[0,:,0:lenss,0:lenss].cpu().numpy()
        score = np.triu(score)

  
        for l, start, end in zip(*np.where(score > threshold)):
            start_id = mapping[start-1][0]
            end_id = mapping[end-1][-1]
            Type = self.id2c[l]
            entities.append(
                #(mapping[start-1][0], mapping[end-1][-1], self.id2c[l])
                {'start_idx': start_id, 'end_idx': end_id, 'type': Type,
                'Chinese_type': self.c2c[Type], 'entity': text[0][0][start_id:end_id+1]}
            )
            
        return entities

    def predict_to_file(self, in_file, out_file):
        """
        Args:
            in_file (str): path of input json file
            out_file (str): path of output json file which can be submitted to CBLUE
        """
        data = json.load(open(in_file))
        for d in data:
            d['entities'] = []
            entities = self.recognize(d['text'])
            #print(entities)
            for e in entities:
                """
                d['entities'].append({
                    'start_idx': e[0],
                    'end_idx': e[1],
                    'type': e[2]
                    #'entity': e[3]
                })
                """
                d['entities'].append({
                    'start_idx': e['start_idx'],
                    'end_idx': e['end_idx'],
                    'type': e['type']
                })
        json.dump(
            data,
            open(out_file, 'w', encoding='utf-8'),
            indent=4,
            ensure_ascii=False
        )

_id2c = {0: 'dis', 1: 'sym', 2: 'pro', 3: 'equ', 4: 'dru', 5: 'ite', 6: 'bod', 7: 'dep', 8: 'mic'}
_c2c = {'dis': "疾病", 'sym': "临床表现", 'pro': "医疗程序", 'equ': "医疗设备", 'dru': "药物", 
        'ite': "医学检验项目", 'bod': "身体", 'dep': "科室", 'mic': "微生物类"}

def load_NNER(model_save_path='./checkpoint/macbert-large_dict.pth', 
                maxlen=512, c_size=9, id2c=_id2c, c2c=_c2c):
    '''
    Args:
        model_save_path (string, optional): path of pretrained model
            Default: './checkpoint/macbert-large_dict.pth'
        maxlen (int, optional): max length of sentence
            Default: 512
        c_size (int, optional): number of entity class
            Default: 9
        id2c (dictionary, optional): mapping between id and entity class
            Default: _id2c
        c2c (dictionary, optional): mapping between entity class and its Chinese meaning
            Default: _c2c
    Returns:
        NNER (class _NamedEntityRecognizer): the pretrained NNER model
    '''
    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    print("Using {} device".format(device))

    model_name = 'hfl/chinese-macbert-large'
    tokenizer = BertTokenizerFast.from_pretrained(model_name)
    tokenizer.add_special_tokens({'pad_token': '[PAD]'})
    model_base = _GlobalPointerNet(model_name).to(device)

    if not os.path.exists(model_save_path):
        model_dir = os.path.dirname(model_save_path)
        os.makedirs(model_dir)
        # Reference:
        # https://github.com/pytorch/pytorch/blob/b5b62b340891f041b378681577c74922d21700a9/torch/hub.py
        url = 'https://github.com/cb1cyf/NNER/releases/download/v0.0.1/macbert-large_dict.pth'
        #model_base = torch.hub.load_state_dict_from_url(url, model_dir)
        torch.hub.download_url_to_file(url, model_save_path)
    state_dict = torch.load(model_save_path, map_location=device)
    # multi-gpu -> 1 gpu
    model_base.load_state_dict({k.replace('module.', ''): v for k, v in state_dict.items()})
    NNER = _NamedEntityRecognizer(maxlen, c_size, id2c, c2c, device, tokenizer, model_base)
    return NNER



import json
import torch
import numpy as np
from torch import nn
from torch.utils.data import Dataset, DataLoader
from transformers import BertModel, BertTokenizerFast
from .model import GlobalPointer
from .utils import rematch

#__all__=['model_save_path','model_base','NamedEntityRecognizer','in_file','out_file','predict_to_file']

#########################################

'''
name_tag = 0 # 0代表使用模型名称导入bert 1代表使用模型路径导入bert
model_path = "./pretrain/roberta_wwm_ext_large"
model_save_path = "./checkpoint/roberta_wwm_ext_large"
'''

model_name = "hfl/chinese-macbert-large"
#model_name = "hfl/chinese-roberta-wwm-ext-large"
model_save_path = "./checkpoint/our_model"


model_save_path ="./checkpoint/68.2796_macbert_large/macbert-large"

out_file = './CMeEE_test_answer.json'
in_file = './dataset/normal/CMeEE_test.json'
train_file = './dataset/normal/CMeEE_train.json'
val_file = './dataset/normal/CMeEE_dev.json'

head_size = 64
embedding_size = 1024 # large:1024 or base:768
maxlen = 512 # len of sentence
epochs = 10
batch_size = 8
learning_rate = 5e-6 # 1e-5
tag_fgm = 1 #0代表不使用对抗训练，1代表使用对抗训练
n_eps=0.1

##########################################



device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

print("Using {} device".format(device))


tokenizer = BertTokenizerFast.from_pretrained(model_name)

tokenizer.add_special_tokens({'pad_token': '[PAD]'})
c_size = 9
c2id = {'dis': 0, 'sym': 1, 'pro': 2, 'equ': 3, 'dru': 4, 'ite': 5, 'bod': 6, 'dep': 7, 'mic': 8}
id2c = {0: 'dis', 1: 'sym', 2: 'pro', 3: 'equ', 4: 'dru', 5: 'ite', 6: 'bod', 7: 'dep', 8: 'mic'}

#gpus = [0, 1, 2, 3]
#torch.cuda.set_device('cuda:{}'.format(gpus[0]))




class FGM:
    def __init__(self, model: nn.Module, eps=n_eps):
        self.model = model
        self.eps = eps
        self.backup = {}

    # only attack word embedding
    def attack(self, emb_name='word_embeddings'):
        for name, param in self.model.named_parameters():
            if param.requires_grad and emb_name in name:
                self.backup[name] = param.data.clone()
                norm = torch.norm(param.grad)
                if norm and not torch.isnan(norm):
                    r_at = self.eps * param.grad / norm
                    param.data.add_(r_at)

    def restore(self, emb_name='word_embeddings'):
        for name, para in self.model.named_parameters():
            if para.requires_grad and emb_name in name:
                assert name in self.backup
                para.data = self.backup[name]

        self.backup = {}


def load_data(filename):
    """
    加载数据
    单条格式：[text, (start, end, label), (start, end, label), ...]，
              意味着text[start:end + 1]是类型为label的实体。
    """
    D = []
    for d in json.load(open(filename,encoding='utf-8')):
        D.append([d['text']])
        for e in d['entities']:
            start, end, label = e['start_idx'], e['end_idx'], e['type']
            if start <= end:
                D[-1].append((start, end, label))
            #categories.add(label)
    return D


def load_test_data(filename):
    """
    加载测试数据，仅有文本
    """
    D = []
    for d in json.load(open(filename,encoding='utf-8')):
        D.append([d['text']])
    return D


train_data = load_data(train_file)
val_data = load_data(val_file)


class GlobalPointerNet(nn.Module):
    def __init__(self, model_name):
        super(GlobalPointerNet, self).__init__()
        self.head = GlobalPointer(c_size, head_size, embedding_size)
        self.bert = BertModel.from_pretrained(model_name)

    def forward(self, input_ids, attention_mask, token_type_ids):
        x1 = self.bert(input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
        x2 = x1.last_hidden_state
        #last_hidden_state (torch.FloatTensor of shape (batch_size, sequence_length, hidden_size))
        # – Sequence of hidden-states at the output of the last layer of the model.

        logits = self.head(x2, mask=attention_mask)
        return logits

class CustomDataset(Dataset):
    def __init__(self, data, tokenizer, maxlen):
        self.data = data
        self.tokenizer = tokenizer
        self.maxlen = maxlen

    @staticmethod
    def find_index(offset_mapping, index):
        for idx, internal in enumerate(offset_mapping[1:]):
            if internal[0] <= index < internal[1]:
                return idx + 1
        return None

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        d = self.data[idx]
        label = torch.zeros((c_size, self.maxlen, self.maxlen))
        enc_context = tokenizer(d[0], return_offsets_mapping=True, max_length=self.maxlen, truncation=True,
                                padding='max_length', return_tensors='pt')
        
        enc_context = {key: enc_context[key][0] for key in enc_context.keys() if enc_context[key].shape[0] == 1}
        for entity_info in d[1:]:
            start, end = entity_info[0], entity_info[1]
            offset_mapping = enc_context['offset_mapping']
            start = self.find_index(offset_mapping, start)
            end = self.find_index(offset_mapping, end)
            if start and end and start < self.maxlen and end < self.maxlen:
                label[c2id[entity_info[2]], start, end] = 1
        return enc_context, label

class CustomDataset4test(Dataset):
    def __init__(self, data, tokenizer, maxlen):
        self.data = data
        self.tokenizer = tokenizer
        self.maxlen = maxlen

    @staticmethod
    def find_index(offset_mapping, index):
        for idx, internal in enumerate(offset_mapping[1:]):
            if internal[0] <= index < internal[1]:
                return idx + 1
        return None

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        d = self.data[idx]
        label = torch.zeros((c_size, self.maxlen, self.maxlen))
        enc_context = tokenizer(d[0], return_offsets_mapping=True, max_length=self.maxlen, truncation=True,
                                padding='max_length', return_tensors='pt')
                                
        enc_context = {key: enc_context[key][0] for key in enc_context.keys() if enc_context[key].shape[0] == 1}

        return enc_context


model = GlobalPointerNet(model_name).to(device)
# model = nn.DataParallel(model.cuda(), device_ids=gpus, output_device=gpus[0])


optimizer = torch.optim.AdamW(model.parameters(), lr=learning_rate)
training_data = CustomDataset(train_data, tokenizer, maxlen)
valing_data = CustomDataset(val_data, tokenizer, maxlen)
train_dataloader = DataLoader(training_data, batch_size=batch_size, shuffle=True)
val_dataloader = DataLoader(valing_data, batch_size=batch_size)





def multilabel_categorical_crossentropy(y_true, y_pred):
    y_pred = (1 - 2 * y_true) * y_pred
    y_pred_neg = y_pred - y_true * 1e12
    y_pred_pos = y_pred - (1 - y_true) * 1e12
    zeros = torch.zeros_like(y_pred[..., :1])
    y_pred_neg = torch.cat([y_pred_neg, zeros], dim=-1)
    y_pred_pos = torch.cat([y_pred_pos, zeros], dim=-1)
    neg_loss = torch.logsumexp(y_pred_neg, dim=-1)
    pos_loss = torch.logsumexp(y_pred_pos, dim=-1)
    return neg_loss + pos_loss


def global_pointer_crossentropy(y_true, y_pred):
    """
    给GlobalPointer设计的交叉熵
    """
    bh = y_pred.shape[0] * y_pred.shape[1]
    y_true = torch.reshape(y_true, (bh, -1))
    y_pred = torch.reshape(y_pred, (bh, -1))
    return torch.mean(multilabel_categorical_crossentropy(y_true, y_pred))


def global_pointer_f1_score(y_true, y_pred):
    y_pred = torch.greater(y_pred, 0)
    return torch.sum(y_true * y_pred).item(), torch.sum(y_true + y_pred).item()


def train(dataloader, model, loss_fn, optimizer, fgm_flag=0):
    if fgm_flag==1:
        model.train()
        size = len(dataloader.dataset)
        numerate, denominator = 0, 0
        fgm = FGM(model)#fgm
        for batch, (data, y) in enumerate(dataloader):
            input_ids = data['input_ids'].to(device)
            attention_mask = data['attention_mask'].to(device)
            token_type_ids = data['token_type_ids'].to(device)
            y = y.to(device)

            pred = model(input_ids, attention_mask, token_type_ids)
            loss = loss_fn(y, pred)

            temp_n, temp_d = global_pointer_f1_score(y, pred)
            numerate += temp_n
            denominator += temp_d
            # Backpropagation
            optimizer.zero_grad()
            loss.backward()

            fgm.attack() #fgm 在embedding上添加对抗扰动
            adv_pred = model(input_ids, attention_mask, token_type_ids)#fgm
            loss_adv = loss_fn(y,adv_pred)#fgm 
            loss_adv.backward() #fgm 反向传播，并在正常的grad基础上，累加对抗训练的梯度
            fgm.restore() #fgm 恢复embedding参数
            
            optimizer.step()
            model.zero_grad()
            if batch % 50 == 0:
                loss, current = loss.item(), batch * len(input_ids)
                print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")
                
        print(f"Train F1: {(2 * numerate / denominator):>4f}")
    else:
        model.train()
        size = len(dataloader.dataset)
        numerate, denominator = 0, 0
        for batch, (data, y) in enumerate(dataloader):
            input_ids = data['input_ids'].to(device)
            attention_mask = data['attention_mask'].to(device)
            token_type_ids = data['token_type_ids'].to(device)
            y = y.to(device)
            pred = model(input_ids, attention_mask, token_type_ids)
            loss = loss_fn(y, pred)
            temp_n, temp_d = global_pointer_f1_score(y, pred)
            numerate += temp_n
            denominator += temp_d
        
            # Backpropagation
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            
            if batch % 50 == 0:
                loss, current = loss.item(), batch * len(input_ids)
                print(f"loss: {loss:>7f}  [{current:>5d}/{size:>5d}]")
                
        print(f"Train F1: {(2 * numerate / denominator):>4f}")
        


def test(dataloader, loss_fn, model):
    size = len(dataloader.dataset)
    model.eval()
    test_loss = 0
    numerate, denominator = 0, 0
    with torch.no_grad():
        for data, y in dataloader:
            input_ids = data['input_ids'].to(device)
            attention_mask = data['attention_mask'].to(device)
            token_type_ids = data['token_type_ids'].to(device)
            y = y.to(device)
            pred = model(input_ids, attention_mask, token_type_ids)
            test_loss += loss_fn(y, pred).item()
            temp_n, temp_d = global_pointer_f1_score(y, pred)
            numerate += temp_n
            denominator += temp_d
    test_loss /= size
    test_f1 = 2 * numerate / denominator
    print(f"Val: F1:{(test_f1):>4f},Avg loss: {test_loss:>8f} \n")
    return test_f1


class NamedEntityRecognizer(object):
    """
    命名实体识别器
    """
    def recognize(self, text, threshold=0):
        
        model_base.eval()
        text = text[0:maxlen]
 
        #tokens = tokenizer.tokenize(text, max_length=maxlen, truncation=True)
        tokens = tokenizer.tokenize(text)
        #print(tokens)
        #text = text[0:maxlen]
        mapping = rematch(tokenizer, text, tokens)
        text = [[str(text)]]
        data = CustomDataset4test(text,tokenizer,maxlen)
        data = DataLoader(data)
        input_ids = []
        attention_mask = []
        token_type_ids = []
        
        for data_1 in data:
            input_ids = data_1['input_ids'].to(device)
            attention_mask = data_1['attention_mask'].to(device)
            token_type_ids = data_1['token_type_ids'].to(device)


        with torch.no_grad():
            scores = model_base(input_ids, attention_mask, token_type_ids)

        entities = []
        lenss = len(mapping)

        
        score = scores[0,:,0:lenss,0:lenss].cpu().numpy()
        score = np.triu(score)

  
        for l, start, end in zip(*np.where(score > threshold)):
            entities.append(
                (mapping[start-1][0], mapping[end-1][-1], id2c[l])
            )
            
        return entities
    



def predict_to_file(in_file, out_file):
    """
    预测到文件
    可以提交到 https://tianchi.aliyun.com/dataset/dataDetail?dataId=95414
    """
    data = json.load(open(in_file))
    for d in data:
        d['entities'] = []
        entities = NER.recognize(d['text'])
        print(entities)
        for e in entities:
            d['entities'].append({
                'start_idx': e[0],
                'end_idx': e[1],
                'type': e[2]
                #'entity': e[3]
            })
    json.dump(
        data,
        open(out_file, 'w', encoding='utf-8'),
        indent=4,
        ensure_ascii=False
    )


if __name__ == '__main__':
    
    #以下为训练过程，已注释
    # max_F1 = 0
    # for t in range(epochs):
    #     print(f"Epoch {t + 1}\n-------------------------------")
    #     train(train_dataloader, model, global_pointer_crossentropy, optimizer, tag_fgm)
    #     F1 = test(val_dataloader, global_pointer_crossentropy, model)
    #     if F1 > max_F1:
    #         max_F1 = F1
    #         print(f"Better F1: {(max_F1):>4f}")
    #         torch.save(model,model_save_path)
    #         print('best model saved')
    # print("Done!")
    # print(f"Best F1: {(max_F1):>4f}")
    
    model_base = torch.load(model_save_path)
    NER = NamedEntityRecognizer()
    predict_to_file(in_file, out_file)
    print("answer saved")



import numpy as np
import unicodedata

def to_array(*args):
    """批量转numpy的array
    """
    results = [np.array(a) for a in args]
    if len(args) == 1:
        return results[0]
    else:
        return results

def is_control(ch):
    """控制类字符判断
    """
    return unicodedata.category(ch) in ('Cc', 'Cf')

def is_special(ch):
    """判断是不是有特殊含义的符号
    """
    return bool(ch) and (ch[0] == '[') and (ch[-1] == ']') 
    #and(ch !='[UNK]')

def stem(token):
    """获取token的“词干”（如果是##开头，则自动去掉##）
    """
    if token[:2] == '##':
        return token[2:]
    else:
        return token

def rematch(tokenizer,text, tokens,):
    """
    给出原始的text和tokenize后的tokens的映射关系
    """
    
    text = text.lower()
    normalized_text, char_mapping = '', []
    for i, ch in enumerate(text):
        ch = ''.join([
            c for c in ch
            if not (ord(c) == 0 or ord(c) == 0xfffd or is_control(c))
        ])
        normalized_text += ch
        char_mapping.extend([i] * len(ch))
    
    i = 0
    j = 0
    text, token_mapping, offset = normalized_text, [], 0
    for token in tokens:
        if is_special(token):
            token_mapping.append([i])
            i = i + 1
        else:
            token = stem(token)
            #print(token)
            start = text[offset:].index(token) + offset
            #print(start)
            end = start + len(token)
            token_mapping.append(char_mapping[start:end])
            offset = end
            j = end - start - 1
            i = i + j + 1

        
    return token_mapping

